import argparse
import os
import torch
import distutils

DEFAULT_CHANNELS = {
    "data": "data"
}


def strtobool(v):
    return bool(distutils.util.strtobool(v))


def bool_argument(
    parser: argparse.ArgumentParser,
    *args,
    **kwargs
):
    parser.add_argument(*args, **kwargs, type=strtobool, const=True, nargs="?")


def train_args_export_model(parser: argparse.ArgumentParser,
                            model_dir='model',
                            export_model=True
                            ):
    model_dir = os.environ.get('SM_MODEL_DIR', model_dir)
    parser.add_argument(
        '--model-dir', type=str,
        default=model_dir,
        help='directory to save final model (default: {})'.format(model_dir))
    bool_argument(
        parser,
        '--export-model',
        default=export_model,
        help='export final model (boolean, default: {})'.format(export_model))


def train_args_hyperparameters(
    parser: argparse.ArgumentParser,
    batch_size=64,
    eval_batch_size=64,
    max_epochs=100,
    learning_rate=0.0003
):
    parser.add_argument("--workers", type=int, default=2,
                        help="number of data loading workers (default: 2)")
    parser.add_argument('--batch-size', type=int, default=batch_size, metavar='N',
                        help='input batch size for training (default: {})'.format(batch_size))
    parser.add_argument('--test-batch-size', type=int, default=eval_batch_size, metavar='N',
                        help='input batch size for testing (default: {})'.format(eval_batch_size))
    parser.add_argument('--max-epochs', type=int, default=max_epochs, metavar='N',
                        help='number of epochs to train (default: {})'.format(max_epochs))
    parser.add_argument('--learning-rate', type=float, default=learning_rate, metavar='LR',
                        help='learning rate (default: {})'.format(learning_rate))
    parser.add_argument('--seed', type=int, default=1, metavar='S',
                        help='random seed (default: 1)')


def train_args_worker(parser: argparse.ArgumentParser,
                      output_dir='output',
                      channels=None,
                      device=None):
    if channels is None:
        channels = DEFAULT_CHANNELS
    if device is None:
        device = "cpu" if not torch.cuda.is_available() else "cuda"
    output_dir = os.environ.get('SM_OUTPUT_DIR', output_dir)
    parser.add_argument("--device", type=str, default=device,
                        help="device to use (default: {})".format(device))
    parser.add_argument('--output-dir', type=str,
                        default=output_dir,
                        help='directory for checkpoints, logs, images, or other output files (default: "{}")'.format(output_dir))
    for channel, default in channels.items():
        key = 'SM_CHANNEL_{}'.format(channel.upper())
        if key in os.environ:
            default = os.environ[key]
        else:
            default = os.path.abspath(os.path.join(__file__, default))
        parser.add_argument('--{}'.format(channel), type=str,  default=default,
                            help="input directory for [{}] channel".format(channel))


def train_args_standard(parser: argparse.ArgumentParser,
                        output_dir='output',
                        batch_size=64,
                        eval_batch_size=64,
                        max_epochs=100,
                        learning_rate=0.0003,
                        channels=None,
                        device=None):
    train_args_hyperparameters(
        parser=parser,
        batch_size=batch_size,
        eval_batch_size=eval_batch_size,
        max_epochs=max_epochs,
        learning_rate=learning_rate
    )
    train_args_worker(
        parser=parser,
        output_dir=output_dir,
        channels=channels,
        device=device
    )

    """
    parser.add_argument('--hosts', type=list,
                        default=json.loads(os.environ.get('SM_HOSTS', '[]')))
    parser.add_argument('--current-host', type=str,
                        default=os.environ.get('SM_CURRENT_HOST', ''))
    parser.add_argument('--num-gpus', type=int,
                        default=os.environ.get('SM_NUM_GPUS', 1))
    parser.add_argument('--log-interval', type=int, default=100, metavar='N',
                        help='how many batches to wait before logging training status')
    parser.add_argument('--backend', type=str, default=None,
                        help='backend for distributed training (tcp, gloo on cpu and gloo, nccl on gpu)')
    """


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    train_args_standard(
        parser=parser,
        output_dir='output')
    train_args_export_model(
        parser=parser,
        model_dir='model'
    )
    parser.parse_args(args=['--help'])




from .spec import RunSpec
import os
from ignite.metrics import RunningAverage
from ignite.engine import Engine, Events
from .util import chain_callbacks, auto_metric, timer_metric, print_logs, save_logs, create_plots
from ignite.contrib.handlers import ProgressBar
from ignite.contrib.handlers.mlflow_logger import OutputHandler, global_step_from_engine
LOGS_FNAME = "logs.tsv"
PLOT_FNAME = "plot.svg"


def build_engine(
    spec: RunSpec,
    output_dir=None,
    trainer=None,
    metric_cls=RunningAverage,
    tag="",
    mlflow_logger=None,
    is_training=None
):
    if spec.plot_event is not None or spec.log_event is not None:
        assert output_dir is not None
        plot_fname = os.path.join(output_dir, "{}-{}".format(tag, PLOT_FNAME))
        logs_fname = os.path.join(output_dir, "{}-{}".format(tag, LOGS_FNAME))
    else:
        plot_fname = None
        logs_fname = None
    # Create engine
    engine = Engine(spec.step)
    if trainer is None:
        # training
        trainer = engine
        if is_training is None:
            is_training = True
    else:
        # evaluation
        if is_training is None:
            is_training = False
    spec.set_defaults(is_training=is_training)

    # Attach metrics
    for name, metric in spec.metrics.items():
        auto_metric(metric, cls=metric_cls).attach(engine, name)
    if spec.enable_timer:
        timer_metric(engine=engine)

    # Progress bar
    if spec.enable_pbar:
        ProgressBar().attach(engine, metric_names=spec.pbar_metrics)

    # Print logs
    if spec.print_event is not None:
        engine.add_event_handler(
            event_name=spec.print_event,
            handler=print_logs,
            trainer=trainer,
            fmt=spec.print_fmt,
            metric_fmt=spec.print_metric_fmt,
            metric_names=spec.print_metrics)

    # Save logs
    if spec.log_event is not None:
        engine.add_event_handler(
            event_name=spec.log_event,
            handler=save_logs,
            fname=logs_fname,
            trainer=trainer,
            metric_names=spec.log_metrics)

    # Plot metrics
    if spec.plot_event is not None:
        # Plots require logs
        assert spec.log_event is not None
        engine.add_event_handler(
            event_name=spec.plot_event,
            handler=create_plots,
            logs_fname=logs_fname,
            plots_fname=plot_fname,
            metric_names=spec.plot_metrics)

    # Optional user callback for additional configuration
    chain_callbacks(
        callbacks=spec.callback,
        engine=engine,
        trainer=trainer)

    if mlflow_logger is not None and spec.log_event is not None:
        mlflow_logger.attach(
            engine,
            log_handler=OutputHandler(
                tag=tag,
                metric_names=spec.log_metrics,
                global_step_transform=global_step_from_engine(trainer)
            ),
            event_name=spec.log_event
        )

    return engine



from .trainer import train
from .spec import RunSpec
from .evaluator import evaluate
from .util import get_value, get_mean_value, tensors_to_device
from .args import train_args_standard, train_args_worker, train_args_hyperparameters, train_args_export_model



from setuptools import find_packages, setup

setup(name='pytorch-igniter',
      version='0.0.8',
      author='Ben Striner',
      url='https://github.com/bstriner/pytorch-igniter',
      install_requires=[
          'pytorch-ignite',
          'pyyaml',
          'mlflow'
      ],
      packages=find_packages())

# python setup.py bdist_wheel sdist && twine upload dist\*


import os
import mlflow
import contextlib
from ignite.contrib.handlers.mlflow_logger import MLflowLogger
import yaml

class NullContext(object):
    def __enter__(self):
        pass
    def __exit__ (self, type, value, tb):
        pass

RUN_FNAME = 'run.yaml'

def mlflow_ctx(
    output_dir=None,
    run_id=None,
    mlflow_enable=True,
    allow_new=True
):
    if mlflow_enable:
        # Check for a run already in progress
        active_run = mlflow.active_run()
        if active_run is not None:
            print("MLflow ID {} active".format(active_run.info.run_id))
            return NullContext(), output_dir
        # Use argument
        if run_id is not None:
            print("MLflow ID {} set".format(run_id))
            return mlflow.start_run(run_id=run_id), output_dir
        # Check environment variable
        if 'MLFLOW_RUN_ID' in os.environ:
            run_id = os.environ['MLFLOW_RUN_ID']
            output_dir = os.path.join(output_dir, run_id)
            print("MLflow ID {} from environment".format(run_id))
            return mlflow.start_run(run_id=run_id), output_dir
        # Check saved run id
        if output_dir is not None:
            run_fname = os.path.join(output_dir, RUN_FNAME)
            if os.path.exists(run_fname):
                with open(run_fname) as f:
                    run_id = yaml.load(f, Loader=yaml.SafeLoader)[
                        'info']['run_id']
                print("MLflow ID {} from run file".format(run_id))
                return mlflow.start_run(run_id=run_id), output_dir
        # New run
        if allow_new:
            print("MLflow new run")
            return mlflow.start_run(run_id=run_id), output_dir
        else:
            raise ValueError("No existing MLflow run found")
    else:
        return  NullContext(), output_dir

def get_mlflow_logger(output_dir, mlflow_enable):
    if mlflow_enable:
        mlflow_logger = MLflowLogger()
        if output_dir is not None:
            run_fname = os.path.join(output_dir, RUN_FNAME)
            if not os.path.exists(run_fname):
                active_run = mlflow.active_run()
                active_run = mlflow.get_run(active_run.info.run_id)
                with open(run_fname, 'w') as f:
                    yaml.dump(active_run.to_dictionary(), f)
        return mlflow_logger
    else:
        return None
